from __future__ import unicode_literals

import json
import ipywidgets as widgets
from traitlets import Unicode

MODULE_NAME = 'mlviswidget'
VERSION = '0.0.1'

class Widget(widgets.DOMWidget):
    _model_module = Unicode(MODULE_NAME).tag(sync=True)
    _model_name = Unicode('WidgetModel').tag(sync=True)
    _model_module_version = Unicode(VERSION).tag(sync=True)
    _view_module = Unicode(MODULE_NAME).tag(sync=True)
    _view_name = Unicode('WidgetView').tag(sync=True)
    _view_module_version = Unicode(VERSION).tag(sync=True)
    props = Unicode('{}').tag(sync=True)


class CommonComponent(Widget):
    def __init__(self, props={}, **kwargs):
        super(CommonComponent, self).__init__()
        # TODO: make explicit exception message for the json input is invalid
        self.props = json.dumps(props)



# Dynamically build component wrappers utilizing the ipywidget
import os, sys, json, re, importlib
from pkg_resources import resource_string
from ast import literal_eval
from traitlets import Unicode
from .widget import CommonComponent
from .widget_ext import Manifold


extensions = {'Manifold': Manifold}


current_module = sys.modules[__name__]


def load_jrequirements():
    return literal_eval(resource_string('mlvis', 'jrequirements.json').decode('utf8'))


# Extract the Python component wrapper names from the jrequirements
def extract_components(jrequirements):
    return [component for component in jrequirements]


jrequirements = load_jrequirements()
components = extract_components(jrequirements)


def init(self, props={}):
    for parent in self.__class__.__bases__:
        parent.__init__(self, props=props)


# Dynamically create module classes
for component in components:
    deps = (extensions[component], ) if component in extensions else (CommonComponent, )
    setattr(current_module,
            component,
            type(component,
                 deps,
                 {
                     '_model_name': Unicode(component + 'WidgetModel').tag(sync=True),
                     '_view_name': Unicode(component + 'WidgetView').tag(sync=True),
                     '__init__': init
                 }))



from .widget_builder import *


def _jupyter_nbextension_paths():
    return [{
        'section': 'notebook',
        'src': 'static',
        'dest': 'mlvis',
        'require': 'mlvis/extension'
    }]



from mlvis.widget import CommonComponent
from traitlets import Unicode, observe
import warnings
import pandas as pd
import numpy as np

class Manifold(CommonComponent):
    segments = Unicode('[]').tag(sync=True)


    def __init__(self, props={}):
        if 'data' not in props:
            raise Exception('data must be specified')

        data = props['data']

        valid, exception = self.validate_data(data, give_exception=True)
        if not valid:
            if isinstance(exception, Warning):
                warnings.warn(exception)
            else:
                raise exception

        # make a shallow copy of the props,
        # props are shared with the manifold instance, clone to prevent unexpected changes
        processed_props = props.copy()
        processed_props['data']['x'] = self.process_x(data['x'])
        processed_props['data']['yPred'] = self.process_y_pred(data['yPred'])
        processed_props['data']['yTrue'] = self.process_y_true(data['yTrue'])
        super(Manifold, self).__init__(processed_props)


    @observe('segments')
    def _observe_bar(self, change):
        print(change['old'])
        print(change['new'])


    def validate_data(self, data, give_exception=False):
        if not isinstance(data, dict):
            return False if not give_exception else (False, Exception('Data must in dict.'))
        for key in ['x', 'yPred', 'yTrue']:
            if key not in data:
                return False if not give_exception else (False, Exception('Missing data attribute ' + key + '.'))
        for key in data:
            if key not in ['x', 'yPred', 'yTrue']:
                return False if not give_exception else (False, UserWarning('Unrecognized data attribute ' + key + '.'))
        for key in ['x', 'yTrue']:
            if not isinstance(data[key], list) and not isinstance(data[key], np.ndarray) and not isinstance(data[key], pd.DataFrame):
                return False if not give_exception else (False, Exception(key + ' must be a list/ndarray/dataframe.'))
        if not isinstance(data['yPred'], list) and not isinstance(data['yPred'], np.ndarray):
            return False if not give_exception else (False, Exception('yPred must be a list/ndarray.'))
        for key in ['x', 'yPred', 'yTrue']:
            if len(data[key]) == 0:
                return False if not give_exception else (False, Exception(key + ' can not be empty.'))
        for l in data['yPred']:
            if not (isinstance(l, list) or isinstance(l, pd.DataFrame)):
                return False if not give_exception else (False, Exception('Every item in yPred must be list/DataFrame.'))
        return True if not give_exception else (True, None)


    def check_is_numerical_feature(self, y_pred):
        if isinstance(y_pred[0], pd.DataFrame):
            return y_pred[0].shape[1] == 1
        else:
            return len(y_pred[0]) == 1


    def process_x(self, x):
        """
        Convert the x data frame into the format manifold recognizes
        :param x: An ndarray/list for the feature list
        :return: A list for the x attribute of the data that with the manifold data format
        """
        if isinstance(x, pd.DataFrame):
            return x.to_dict('records')
        elif isinstance(x, np.ndarray):
            return x.tolist()
        else:
            return x


    def process_y_pred(self, y_pred):
        """
        Convert y pred feature -- the predictions of each features for each model into the manifold recognized informat
        :param y_pred: A ndarray/list of data frames. Each data frame is the predict probability of one model,
            which each column being the predicted probability for one class
        :return: A list for the y_pred attribute with the manifold data format
        """
        if isinstance(y_pred, np.ndarray):
            return [y.to_dict('records') if isinstance(y, pd.DataFrame) else y for y in np.nditer(y_pred)]
        else:
            return [y.to_dict('records') if isinstance(y, pd.DataFrame) else y for y in y_pred]


    def process_y_true(self, y_true):
        """
        Convert y true feature into the format manifold recognizes
        :param y_true: An ndarray/list for the ground truth
        :return: A list of the ground truths with the manifold data format
        """
        if isinstance(y_true, pd.DataFrame):
            return y_true.iloc[:, 0].tolist()
        elif isinstance(y_true, np.ndarray):
            return y_true.tolist()
        else:
            return y_true



from setuptools import setup, find_packages, Command
from setuptools.command.develop import develop as _develop
from setuptools.command.install import install as _install
import os, subprocess, json


BASE_DIR = os.path.dirname(os.path.abspath(__file__))


EXTENSION_DIR = os.path.join(os.path.dirname(__file__), "mlvis", "static")


INSTALL_REQUIRES = [
    'traitlets',
    'ipywidgets',
    'pandas'
]


DEFAULT_JREQUIREMENTS = {
    "FeatureListView": {
        "module": {"@mlvis/feature-list-view": "external"}
    },
    "GraphBuilder": {
        "module": {"@mlvis/jupyter-graph-builder": "external"}
    },
    "Manifold": {
        "module": {"@mlvis/juypter-manifold": "external"}
    },
    "MultiWayPlot": {
        "module": {"@mlvis/jupyter-multi-way-plot": "external"}
    },
    "StackedCalendar": {
        "module": {"@mlvis/stacked-calendar": "external"}
    },
    "MACausal": {
        "module": {"@mlvis/jupyter-ma-causal": "external"}
    }
}


def read(*parts):
    return open(os.path.join(BASE_DIR, *parts), 'r').read()


def init_jrequirements():
    try:
        file = open(os.path.join(BASE_DIR, 'mlvis', 'jrequirements.json'), mode='w')
        json.dump(DEFAULT_JREQUIREMENTS, file, indent=2)
    finally:
        if file:
            file.close()


class init_jrequirements_command(Command):
    description = 'customize jrequirements.json'
    user_options = []

    def initialize_options(self): pass
    def finalize_options(self): pass

    def run(self):
        init_jrequirements()


class develop(_develop):
    def run(self):
        from notebook.nbextensions import install_nbextension
        from notebook.services.config import ConfigManager

        _develop.run(self)

        install_nbextension(EXTENSION_DIR, symlink=True,
                            overwrite=True, user=False, destination="mlvis")
        cm = ConfigManager()
        cm.update('notebook', {"load_extensions":
            {
                "mlvis/index": True,
                "mlvis/extension": True
            }
        })


class install(_install):
    def run(self):
        _install.run(self)

        # A hack for installing the install_requires as there seems to be
        # a issue with custom install command:
        # https://github.com/pypa/setuptools/issues/456
        subprocess.check_call(['pip', 'install'] + INSTALL_REQUIRES)


setup(name='mlvis',
      cmdclass={'develop': develop,
                'init_jrequirements': init_jrequirements_command},
      version='0.1.0',
      description='A wrapper around react components for use in jupyter notebooks',
      long_description='{}'.format(read(os.path.join('docs', 'introduction.md'))),
      long_description_content_type='text/markdown',
      keywords=['data', 'visualization', 'machine learning'],
      url='https://github.com/uber/manifold.git',
      author='Hong Wang',
      author_email='hongw@uber.com',
      license='Apache License 2.0',
      include_package_data=True,
      packages=find_packages(),
      zip_safe=False,
      package_data={'': ['mlvis/jrequirements.json']},
      data_files=[
        ('share/jupyter/nbextensions/mlvis', [
            'mlvis/static/extension.js',
            'mlvis/static/index.js',
            'mlvis/static/index.js.map'
        ])
      ],
      scripts=['bin/jpip'],
      install_requires=INSTALL_REQUIRES,
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Multimedia :: Graphics',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Framework :: Jupyter'
        ]
      )
